/*
 * compiz-transition.cpp
 *
 * A small executable to be run at first launch which moves
 * old settings paths to new ones so that the user's settings
 * are preserved
 *
 * Copyright (C) 2010 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Sam Spilsbury <sam.spilsbury@canonical.com>
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <boost/filesystem.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/cerrno.hpp>
#include <boost/foreach.hpp>
#include <cstring>
#include <cstdio>
#include <list>

using namespace std;
namespace bf = boost::filesystem;

bool move_internal_dirs (void)
{
    char *homedir = getenv ("HOME");
    char old_dir[512], new_dir[512];
    char command [1600];

    snprintf (old_dir, sizeof (old_dir), "%s/.compiz", homedir);
    snprintf (new_dir, sizeof (new_dir), "%s/.compiz-1", homedir);

    snprintf (command, sizeof (command), "cp -a %s %s", old_dir, new_dir);

    return true;
}

bool
move_ini_setting (list <string> & lines, string setting, string old_plugin, string new_plugin)
{
    list <string>::iterator  old_plugin_begin (lines.end ());
    list <string>::iterator  old_plugin_end (lines.end ());
    list <string>::iterator  new_plugin_begin (lines.end ());
    list <string>::iterator  new_plugin_end (lines.end ());
    string ini_setting_name;

    for (list<string>::iterator it = lines.begin (); it != lines.end ();
	 it++)
    {
	if (it->find ("[" + old_plugin + "]") != string::npos &&
	    old_plugin_begin == lines.end ())
	{
	    old_plugin_begin = it;
	    continue;
	}

	if (old_plugin_begin != lines.end () && it->find ("[") != string::npos)
	{
	    old_plugin_end = it;
	    break;
	}
    }

    /* Couldn't find old_plugin, so just return false */
    if (old_plugin_begin == lines.end ())
    {
	cout << "[LOG]: Could not find plugin " << old_plugin << " settings, skipping" << endl;
	return false;
    }

    for (list<string>::iterator it = lines.begin (); it != lines.end ();
	 it++)
    {
	if (it->find ("[" + new_plugin + "]") != string::npos &&
	    new_plugin_begin == lines.end ())
	{
	    new_plugin_begin = it;
	    continue;
	}

	if (new_plugin_begin != lines.end () && it->find ("[") != string::npos)
	{
	    new_plugin_end = it;
	    break;
	}
    }

    /* No new plugin section found in the ini file, so create it */
    if (new_plugin_begin == lines.end ())
    {
	cout << "[LOG]: Adding new plugin [" << new_plugin << "] to config" << endl;
	lines.push_back ("");
	lines.push_back ("[" + new_plugin + "]");
	new_plugin_begin = lines.end ();
	new_plugin_end = lines.end ();
    }

    for (list <string>::iterator it = old_plugin_begin; it != old_plugin_end; it++)
    {
	if (it->find (setting) != string::npos)
	{
	    ini_setting_name = (*it);
	    lines.erase (it);
	    break;
	}
    }

    lines.insert (new_plugin_end, ini_setting_name);
}

bool convert_ini_file (bf::path ini_file_path)
{
    ifstream fp;
    ofstream ofp;
    char     *buffer;
    int	     length;
    string   data;
    list <string> lines;

    cout << "[LOG]: Opening file " << ini_file_path.file_string () << endl;

    fp.open (ini_file_path.file_string ().c_str ());

    if (!fp.is_open ())
    {
	cout << "[WARNING]: Unable to open file " << ini_file_path.file_string () << endl;
	return false;
    }

    fp.seekg (0, ios::end);
    length = fp.tellg ();
    length++;
    fp.seekg (0, ios::beg);

    buffer = new char [length];

    fp.read (buffer, length - 1);
    buffer[length -1] = '\0';
    fp.close ();

    data = buffer;

    delete[] buffer;

    boost::split (lines, data, boost::is_any_of ("\n"));

    cout << "[LOG] 0.8.6 configuration file data is: " << endl << endl;

    BOOST_FOREACH (string &l, lines)
    {
	cout << l << endl;
    }

    BOOST_FOREACH (string &l, lines)
    {
	size_t pos = 0;

	/* Find and replace decoration with decor */
	if ((pos = l.find ("decoration", pos)) != string::npos)
	{
	    cout << "[LOG]: Transitioning decoration -> decor" << endl;
	    l.erase (pos, 10);
	    l.insert (pos, "decor");
	}

	pos = 0;

	/* Find and replace png, svg, jpeg with img* */
	if ((pos = l.find (";png", pos)) != string::npos)
	{
	    cout << "[LOG]: Transitioning png -> imgpng" << endl;
	    l.erase (pos, 4);
	    l.insert (pos, ";imgpng");
	}

	pos = 0;

	if ((pos = l.find (";svg", pos)) != string::npos)
	{
	    cout << "[LOG]: Transitioning svg -> imgsvg" << endl;
	    l.erase (pos, 4);
	    l.insert (pos, ";imgsvg");
	}

	pos = 0;

	if ((pos = l.find (";jpeg", pos)) != string::npos)
	{
	    cout << "[LOG]: Transitioning decoration -> decor" << endl;
	    l.erase (pos, 4);
	    l.insert (pos, ";imgjpeg");
	}

	pos = 0;

	/* Find and replace as_ with s0_ */
	if ((pos = l.find ("as_", pos)) != string::npos)
	{
	    cout << "[LOG]: Transitioning display level setting " << l << " to screen one" << endl;
	    l.erase (pos, 3);
	    l.insert (pos, "s0_");
	}

	pos = 0;

	/* Insert new plugins that we will load automatically */
	if ((pos = l.find ("= core;", pos)) != string::npos)
	{
	    l.insert (pos + 7, "composite;opengl;compiztoolbox;");
	    cout << "[LOG]: New plugins list is " << l << endl;
	}

	pos = 0;

    }

    move_ini_setting (lines, "slow_animations_key", "core", "composite");
    move_ini_setting (lines, "detect_refresh_rate", "core", "composite");
    move_ini_setting (lines, "refresh_rate", "core", "composite");
    move_ini_setting (lines, "unredirect_fullscreen_windows", "core", "composite");
    move_ini_setting (lines, "force_independent_output_painting", "core", "composite");
    move_ini_setting (lines, "texture_compression", "core", "opengl");
    move_ini_setting (lines, "texture_filter", "core", "opengl");
    move_ini_setting (lines, "lighting", "core", "opengl");
    move_ini_setting (lines, "sync_to_vblank", "core", "opengl");

    cout << "[LOG] 0.9.2.1 configuration file data is: " << endl;

    BOOST_FOREACH (string &line, lines)
    {
	cout << line << endl;
    }

    cout << "[LOG]: Writing configuration data" << endl;

    ofp.open (ini_file_path.file_string ().c_str ());

    if (!ofp.is_open ())
    {
	cout << "[LOG]: Error opening " << ini_file_path.file_string () << " for writing!";
	return false;
    }
    else
    {
	BOOST_FOREACH (string &line, lines)
	{
	    ofp << endl;
	    ofp << line;
	}
    }

    ofp.close ();

    return true;
}

bool move_ini_files ()
{
    char *homedir = getenv ("HOME");
    char old_dir[512], new_dir[512];
    char command[1600];
    boost::system::error_code ec;

    snprintf (old_dir, sizeof (old_dir), "%s/.config/compiz", homedir);
    snprintf (new_dir, sizeof (new_dir), "%s/.config/compiz-1", homedir);

    bf::path old_dir_p (bf::system_complete (bf::path (old_dir)));
    bf::path new_dir_p (bf::system_complete (bf::path (new_dir)));
    bf::path profile_dir_p;

    snprintf (command, sizeof (command), "cp -a %s %s", old_dir, new_dir);

    system (command);

    profile_dir_p = new_dir_p.file_string () + "/compizconfig";

    if (bf::is_directory (profile_dir_p))
    {
	bf::directory_iterator end_it;
	cout << "[LOG]: Scanning directory " << profile_dir_p.file_string () << " for configuration files" << endl;

	for (bf::directory_iterator it (profile_dir_p); it != end_it; ++it)
	{
	    if (it->path ().file_string () == "config")
		continue;
	    else if (it->path ().file_string ().find (".ini") != string::npos)
	    {
		cout << "[LOG]: Converting profile called " << it->path ().file_string ();

		if (!convert_ini_file (it->path ()))
		    cout << "[WARNING]: Unable to convert profile " << it->path ().file_string () << endl;
		else
		    cout << "[LOG]: Successfully converted profile " << it->path ().file_string () << endl;
	    }
	}
    }
    else
    {
	cout << "[WARNING]: Could not find profiles directory!" << endl;
	return false;
    }

    return true;
}
    

int main (void)
{
    cout << "Compiz Migration Script for Ubuntu 11.04" << endl;
    cout << "Moving settings from Compiz 0.8.6 to Compiz 0.9.4" << endl;

    cout << "[LOG]: Moving Internal Files" << endl;

    if (move_internal_dirs ())
	cout << "[LOG]: Successfully moved internal files" << endl;
    else
	cout << "[WARNING]: Could not move internal files" << endl;

    cout << "[LOG]: Moving and rewriting INI configuration files" << endl;

    if (move_ini_files ())
	cout << "[LOG]: Successfully moved INI configuration files" << endl;
    else
	cout << "[WARNING]: Could not rewrite and move all INI configuration files, your settings may be reset" << endl;

    return 0;
}
